@@ -112,11 +112,6 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 		flags |= CLIENT_MULTI_RESULTS; /* needed for mysql_multi_query() */
 		/* remove some insecure options */
 		flags &= ~CLIENT_MULTI_STATEMENTS;   /* don't allow multi_queries via connect parameter */
-#ifndef MYSQLI_USE_MYSQLND
-		if (PG(open_basedir) && PG(open_basedir)[0] != '\0') {
-			flags &= ~CLIENT_LOCAL_FILES;
-		}
-#endif
 	}
 
 	if (!socket_len || !socket) {
@@ -179,9 +174,7 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 #else
 							if (!mysql_ping(mysql->mysql)) {
 #endif
-#ifdef MYSQLI_USE_MYSQLND
 								mysqlnd_restart_psession(mysql->mysql);
-#endif
 								MyG(num_active_persistent)++;
 
 								/* clear error */
@@ -189,7 +182,6 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 
 								goto end;
 							} else {
-#ifdef MYSQLI_USE_MYSQLND
 								if (mysql->mysql->data->vio->data->ssl) {
 									/* copy over pre-existing ssl settings so we can reuse them when reconnecting */
 									ssl = true;
@@ -200,22 +192,7 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 									ssl_capath = mysql->mysql->data->vio->data->options.ssl_capath ? estrdup(mysql->mysql->data->vio->data->options.ssl_capath) : NULL;
 									ssl_cipher = mysql->mysql->data->vio->data->options.ssl_cipher ? estrdup(mysql->mysql->data->vio->data->options.ssl_cipher) : NULL;
 								}
-#else
-								if (mysql->mysql->options.ssl_key
-										|| mysql->mysql->options.ssl_cert
-										|| mysql->mysql->options.ssl_ca
-										|| mysql->mysql->options.ssl_capath
-										|| mysql->mysql->options.ssl_cipher) {
-									/* copy over pre-existing ssl settings so we can reuse them when reconnecting */
-									ssl = true;
 
-									ssl_key = mysql->mysql->options.ssl_key ? estrdup(mysql->mysql->options.ssl_key) : NULL;
-									ssl_cert = mysql->mysql->options.ssl_cert ? estrdup(mysql->mysql->options.ssl_cert) : NULL;
-									ssl_ca = mysql->mysql->options.ssl_ca ? estrdup(mysql->mysql->options.ssl_ca) : NULL;
-									ssl_capath = mysql->mysql->options.ssl_capath ? estrdup(mysql->mysql->options.ssl_capath) : NULL;
-									ssl_cipher = mysql->mysql->options.ssl_cipher ? estrdup(mysql->mysql->options.ssl_cipher) : NULL;
-								}
-#endif
 								mysqli_close(mysql->mysql, MYSQLI_CLOSE_IMPLICIT);
 								mysql->mysql = NULL;
 							}
@@ -243,46 +220,12 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 		goto err;
 	}
 	if (!mysql->mysql) {
-#ifndef MYSQLI_USE_MYSQLND
-		if (!(mysql->mysql = mysql_init(NULL))) {
-#else
 		if (!(mysql->mysql = mysqlnd_init(MYSQLND_CLIENT_NO_FLAG, persistent))) {
-#endif
 			goto err;
 		}
 		new_connection = true;
 	}
 
-#ifndef MYSQLI_USE_MYSQLND
-	/* BC for prior to bug fix #53425 */
-	flags |= CLIENT_MULTI_RESULTS;
-
-	if (ssl) {
-		/* if we're here, this means previous conn was ssl, repopulate settings */
-		mysql_ssl_set(mysql->mysql, ssl_key, ssl_cert, ssl_ca, ssl_capath, ssl_cipher);
-
-		if (ssl_key) {
-		    efree(ssl_key);
-		}
-
-		if (ssl_cert) {
-		    efree(ssl_cert);
-		}
-
-		if (ssl_ca) {
-		    efree(ssl_ca);
-		}
-
-		if (ssl_capath) {
-		    efree(ssl_capath);
-		}
-
-		if (ssl_cipher) {
-		    efree(ssl_cipher);
-		}
-	}
-	if (mysql_real_connect(mysql->mysql, hostname, username, passwd, dbname, port, socket, flags) == NULL)
-#else
 	if (ssl) {
 		/* if we're here, this means previous conn was ssl, repopulate settings */
 		mysql_ssl_set(mysql->mysql, ssl_key, ssl_cert, ssl_ca, ssl_capath, ssl_cipher);
@@ -309,7 +252,6 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 	}
 	if (mysqlnd_connect(mysql->mysql, hostname, username, passwd, passwd_len, dbname, dbname_len,
 						port, socket, flags, MYSQLND_CLIENT_NO_FLAG) == NULL)
-#endif
 	{
 		/* Save error messages - for mysqli_connect_error() & mysqli_connect_errno() */
 		php_mysqli_set_error(mysql_errno(mysql->mysql), (char *) mysql_error(mysql->mysql));
@@ -326,18 +268,12 @@ void mysqli_common_connect(INTERNAL_FUNCTION_PARAMETERS, bool is_real_connect, b
 	/* clear error */
 	php_mysqli_set_error(mysql_errno(mysql->mysql), (char *) mysql_error(mysql->mysql));
 
-#ifndef MYSQLI_USE_MYSQLND
-	char reconnect = MyG(reconnect);
-	mysql_options(mysql->mysql, MYSQL_OPT_RECONNECT, (char *)&reconnect);
-#endif
 	unsigned int allow_local_infile = MyG(allow_local_infile);
 	mysql_options(mysql->mysql, MYSQL_OPT_LOCAL_INFILE, (char *)&allow_local_infile);
 
-#if MYSQL_VERSION_ID >= 80021 || defined(MYSQLI_USE_MYSQLND)
 	if (MyG(local_infile_directory) && !php_check_open_basedir(MyG(local_infile_directory))) {
 		mysql_options(mysql->mysql, MYSQL_OPT_LOAD_DATA_LOCAL_DIR, MyG(local_infile_directory));
 	}
-#endif
 
 end:
 	if (!mysqli_resource) {
@@ -500,7 +436,6 @@ PHP_FUNCTION(mysqli_fetch_all)
 }
 /* }}} */
 
-#ifdef MYSQLI_USE_MYSQLND
 /* {{{ Returns statistics about the zval cache */
 PHP_FUNCTION(mysqli_get_client_stats)
 {
@@ -525,7 +460,6 @@ PHP_FUNCTION(mysqli_get_connection_stats)
 
 	mysqlnd_get_connection_stats(mysql->mysql, return_value);
 }
-#endif
 /* }}} */
 
 /* {{{ Fetches all client errors */
@@ -538,38 +472,21 @@ PHP_FUNCTION(mysqli_error_list)
 		RETURN_THROWS();
 	}
 	MYSQLI_FETCH_RESOURCE_CONN(mysql, mysql_link, MYSQLI_STATUS_VALID);
-#ifdef MYSQLI_USE_MYSQLND
-	if (1) {
-		MYSQLND_ERROR_LIST_ELEMENT * message;
-		zend_llist_position pos;
-		array_init(return_value);
-		for (message = (MYSQLND_ERROR_LIST_ELEMENT *) zend_llist_get_first_ex(&mysql->mysql->data->error_info->error_list, &pos);
-			 message;
-			 message = (MYSQLND_ERROR_LIST_ELEMENT *) zend_llist_get_next_ex(&mysql->mysql->data->error_info->error_list, &pos))
-		{
-			zval single_error;
-			array_init(&single_error);
-			add_assoc_long_ex(&single_error, "errno", sizeof("errno") - 1, message->error_no);
-			add_assoc_string_ex(&single_error, "sqlstate", sizeof("sqlstate") - 1, message->sqlstate);
-			add_assoc_string_ex(&single_error, "error", sizeof("error") - 1, message->error);
-			add_next_index_zval(return_value, &single_error);
-		}
-	} else {
-		RETURN_EMPTY_ARRAY();
-	}
-#else
-	if (mysql_errno(mysql->mysql)) {
+
+	MYSQLND_ERROR_LIST_ELEMENT * message;
+	zend_llist_position pos;
+	array_init(return_value);
+	for (message = (MYSQLND_ERROR_LIST_ELEMENT *) zend_llist_get_first_ex(&mysql->mysql->data->error_info->error_list, &pos);
+			message;
+			message = (MYSQLND_ERROR_LIST_ELEMENT *) zend_llist_get_next_ex(&mysql->mysql->data->error_info->error_list, &pos))
+	{
 		zval single_error;
-		array_init(return_value);
 		array_init(&single_error);
-		add_assoc_long_ex(&single_error, "errno", sizeof("errno") - 1, mysql_errno(mysql->mysql));
-		add_assoc_string_ex(&single_error, "sqlstate", sizeof("sqlstate") - 1, mysql_sqlstate(mysql->mysql));
-		add_assoc_string_ex(&single_error, "error", sizeof("error") - 1, mysql_error(mysql->mysql));
+		add_assoc_long_ex(&single_error, "errno", sizeof("errno") - 1, message->error_no);
+		add_assoc_string_ex(&single_error, "sqlstate", sizeof("sqlstate") - 1, message->sqlstate);
+		add_assoc_string_ex(&single_error, "error", sizeof("error") - 1, message->error);
 		add_next_index_zval(return_value, &single_error);
-	} else {
-		RETURN_EMPTY_ARRAY();
 	}
-#endif
 }
 /* }}} */
 
@@ -583,7 +500,6 @@ PHP_FUNCTION(mysqli_stmt_error_list)
 		RETURN_THROWS();
 	}
 	MYSQLI_FETCH_RESOURCE_STMT(stmt, mysql_stmt, MYSQLI_STATUS_INITIALIZED);
-#ifdef MYSQLI_USE_MYSQLND
 	if (stmt->stmt && stmt->stmt->data && stmt->stmt->data->error_info) {
 		MYSQLND_ERROR_LIST_ELEMENT * message;
 		zend_llist_position pos;
@@ -602,19 +518,6 @@ PHP_FUNCTION(mysqli_stmt_error_list)
 	} else {
 		RETURN_EMPTY_ARRAY();
 	}
-#else
-	if (mysql_stmt_errno(stmt->stmt)) {
-		zval single_error;
-		array_init(return_value);
-		array_init(&single_error);
-		add_assoc_long_ex(&single_error, "errno", sizeof("errno") - 1, mysql_stmt_errno(stmt->stmt));
-		add_assoc_string_ex(&single_error, "sqlstate", sizeof("sqlstate") - 1, mysql_stmt_sqlstate(stmt->stmt));
-		add_assoc_string_ex(&single_error, "error", sizeof("error") - 1, mysql_stmt_error(stmt->stmt));
-		add_next_index_zval(return_value, &single_error);
-	} else {
-		RETURN_EMPTY_ARRAY();
-	}
-#endif
 }
 /* }}} */
 
@@ -640,32 +543,15 @@ PHP_FUNCTION(mysqli_multi_query)
 
 	MYSQLI_ENABLE_MQ;
 	if (mysql_real_query(mysql->mysql, query, query_len)) {
-#ifndef MYSQLI_USE_MYSQLND
-		char s_error[MYSQL_ERRMSG_SIZE], s_sqlstate[SQLSTATE_LENGTH+1];
-		unsigned int s_errno;
-		/* we have to save error information, cause
-		MYSQLI_DISABLE_MQ will reset error information */
-		strcpy(s_error, mysql_error(mysql->mysql));
-		strcpy(s_sqlstate, mysql_sqlstate(mysql->mysql));
-		s_errno = mysql_errno(mysql->mysql);
-#else
 		MYSQLND_ERROR_INFO error_info = *mysql->mysql->data->error_info;
 		mysql->mysql->data->error_info->error_list.head = NULL;
 		mysql->mysql->data->error_info->error_list.tail = NULL;
 		mysql->mysql->data->error_info->error_list.count = 0;
-#endif
 		MYSQLI_REPORT_MYSQL_ERROR(mysql->mysql);
 		MYSQLI_DISABLE_MQ;
 
-#ifndef MYSQLI_USE_MYSQLND
-		/* restore error information */
-		strcpy(mysql->mysql->net.last_error, s_error);
-		strcpy(mysql->mysql->net.sqlstate, s_sqlstate);
-		mysql->mysql->net.last_errno = s_errno;
-#else
 		zend_llist_clean(&mysql->mysql->data->error_info->error_list);
 		*mysql->mysql->data->error_info = error_info;
-#endif
 		RETURN_FALSE;
 	}
 	RETURN_TRUE;
@@ -694,11 +580,7 @@ PHP_FUNCTION(mysqli_query)
 	if ((resultmode & ~MYSQLI_ASYNC) != MYSQLI_USE_RESULT &&
 		MYSQLI_STORE_RESULT != (resultmode & ~(MYSQLI_ASYNC | MYSQLI_STORE_RESULT_COPY_DATA))
 	) {
-		zend_argument_value_error(ERROR_ARG_POS(3), "must be either MYSQLI_USE_RESULT or MYSQLI_STORE_RESULT"
-			#ifdef MYSQLI_USE_MYSQLND
-				" with MYSQLI_ASYNC as an optional bitmask flag"
-			#endif
-		);
+		zend_argument_value_error(ERROR_ARG_POS(3), "must be either MYSQLI_USE_RESULT or MYSQLI_STORE_RESULT with MYSQLI_ASYNC as an optional bitmask flag");
 		RETURN_THROWS();
 	}
 
@@ -707,7 +589,6 @@ PHP_FUNCTION(mysqli_query)
 	MYSQLI_DISABLE_MQ;
 
 
-#ifdef MYSQLI_USE_MYSQLND
 	if (resultmode & MYSQLI_ASYNC) {
 		if (mysqli_async_query(mysql->mysql, query, query_len)) {
 			MYSQLI_REPORT_MYSQL_ERROR(mysql->mysql);
@@ -716,7 +597,6 @@ PHP_FUNCTION(mysqli_query)
 		mysql->async_result_fetch_type = resultmode & ~MYSQLI_ASYNC;
 		RETURN_TRUE;
 	}
-#endif
 
 	if (mysql_real_query(mysql->mysql, query, query_len)) {
 		MYSQLI_REPORT_MYSQL_ERROR(mysql->mysql);
@@ -731,11 +611,7 @@ PHP_FUNCTION(mysqli_query)
 		RETURN_TRUE;
 	}
 
-#ifdef MYSQLI_USE_MYSQLND
 	switch (resultmode & ~(MYSQLI_ASYNC | MYSQLI_STORE_RESULT_COPY_DATA)) {
-#else
-	switch (resultmode & ~MYSQLI_ASYNC) {
-#endif
 		case MYSQLI_STORE_RESULT:
 			result = mysql_store_result(mysql->mysql);
 			break;
@@ -744,8 +620,7 @@ PHP_FUNCTION(mysqli_query)
 			break;
 	}
 	if (!result) {
-		php_mysqli_throw_sql_exception((char *)mysql_sqlstate(mysql->mysql), mysql_errno(mysql->mysql),
-										"%s", mysql_error(mysql->mysql));
+		MYSQLI_REPORT_MYSQL_ERROR(mysql->mysql);
 		RETURN_FALSE;
 	}
 
@@ -760,7 +635,6 @@ PHP_FUNCTION(mysqli_query)
 }
 /* }}} */
 
-#ifdef MYSQLI_USE_MYSQLND
 #include "php_network.h"
 /* {{{ mysqlnd_zval_array_to_mysqlnd_array functions */
 static int mysqlnd_zval_array_to_mysqlnd_array(zval *in_array, MYSQLND ***out_array)
@@ -776,7 +650,7 @@ static int mysqlnd_zval_array_to_mysqlnd_array(zval *in_array, MYSQLND ***out_ar
 		i++;
 		if (Z_TYPE_P(elem) != IS_OBJECT ||
 			!instanceof_function(Z_OBJCE_P(elem), mysqli_link_class_entry)) {
-			zend_argument_type_error(i, "must be an instance of mysqli, %s given", zend_zval_type_name(elem));
+			zend_argument_type_error(i, "must be an instance of mysqli, %s given", zend_zval_value_name(elem));
 			return FAILURE;
 		} else {
 			MY_MYSQL *mysql;
@@ -799,12 +673,11 @@ static int mysqlnd_zval_array_to_mysqlnd_array(zval *in_array, MYSQLND ***out_ar
 /* }}} */
 
 /* {{{ mysqlnd_zval_array_from_mysqlnd_array */
-static int mysqlnd_zval_array_from_mysqlnd_array(MYSQLND **in_array, zval *out_array)
+static zend_result mysqlnd_zval_array_from_mysqlnd_array(MYSQLND **in_array, zval *out_array)
 {
 	MYSQLND **p = in_array;
 	zval dest_array;
 	zval *elem, *dest_elem;
-	int ret = 0;
 
 	array_init_size(&dest_array, zend_hash_num_elements(Z_ARRVAL_P(out_array)));
 
@@ -827,7 +700,6 @@ static int mysqlnd_zval_array_from_mysqlnd_array(MYSQLND **in_array, zval *out_a
 				if (dest_elem) {
 					zval_add_ref(dest_elem);
 				}
-				ret++;
 				p++;
 			}
 		}
@@ -837,16 +709,15 @@ static int mysqlnd_zval_array_from_mysqlnd_array(MYSQLND **in_array, zval *out_a
 	zval_ptr_dtor(out_array);
 	ZVAL_COPY_VALUE(out_array, &dest_array);
 
-	return 0;
+	return SUCCESS;
 }
 /* }}} */
 
 /* {{{ mysqlnd_dont_poll_zval_array_from_mysqlnd_array */
-static int mysqlnd_dont_poll_zval_array_from_mysqlnd_array(MYSQLND **in_array, zval *in_zval_array, zval *out_array)
+static void mysqlnd_dont_poll_zval_array_from_mysqlnd_array(MYSQLND **in_array, zval *in_zval_array, zval *out_array)
 {
 	MYSQLND **p = in_array;
 	zval proxy, *elem, *dest_elem;
-	int ret = 0;
 
 	array_init(&proxy);
 	if (in_array) {
@@ -859,7 +730,6 @@ static int mysqlnd_dont_poll_zval_array_from_mysqlnd_array(MYSQLND **in_array, z
 				if (dest_elem) {
 					zval_add_ref(dest_elem);
 				}
-				ret++;
 				p++;
 			}
 		} ZEND_HASH_FOREACH_END();
@@ -868,8 +738,6 @@ static int mysqlnd_dont_poll_zval_array_from_mysqlnd_array(MYSQLND **in_array, z
 	/* destroy old array and add new one */
 	zval_ptr_dtor(out_array);
 	ZVAL_COPY_VALUE(out_array, &proxy);
-
-	return 0;
 }
 /* }}} */
 
@@ -894,10 +762,9 @@ PHP_FUNCTION(mysqli_poll)
 		RETURN_THROWS();
 	}
 
-	// TODO Error promotion
 	if (!r_array && !e_array) {
-		php_error_docref(NULL, E_WARNING, "No stream arrays were passed");
-		RETURN_FALSE;
+		zend_value_error("No stream arrays were passed");
+		RETURN_THROWS();
 	}
 
 	if (r_array != NULL) {
@@ -1018,7 +885,6 @@ PHP_FUNCTION(mysqli_stmt_get_result)
 	MYSQLI_RETVAL_RESOURCE(mysqli_resource, mysqli_result_class_entry);
 }
 /* }}} */
-#endif
 
 /* {{{ */
 PHP_FUNCTION(mysqli_get_warnings)
@@ -1034,11 +900,7 @@ PHP_FUNCTION(mysqli_get_warnings)
 	MYSQLI_FETCH_RESOURCE_CONN(mysql, mysql_link, MYSQLI_STATUS_VALID);
 
 	if (mysql_warning_count(mysql->mysql)) {
-#ifdef MYSQLI_USE_MYSQLND
 		w = php_get_warnings(mysql->mysql->data);
-#else
-		w = php_get_warnings(mysql->mysql);
-#endif
 	}
 	if (!w) {
 		RETURN_FALSE;
@@ -1104,11 +966,7 @@ PHP_FUNCTION(mysqli_get_charset)
 	zval					*mysql_link;
 	const char 				*name = NULL, *collation = NULL, *dir = NULL, *comment = NULL;
 	uint32_t				minlength, maxlength, number, state;
-#ifndef MYSQLI_USE_MYSQLND
-	MY_CHARSET_INFO			cs;
-#else
 	const MYSQLND_CHARSET	*cs;
-#endif
 
 	if (zend_parse_method_parameters(ZEND_NUM_ARGS(), getThis(), "O", &mysql_link, mysqli_link_class_entry) == FAILURE) {
 		RETURN_THROWS();
@@ -1116,17 +974,6 @@ PHP_FUNCTION(mysqli_get_charset)
 	MYSQLI_FETCH_RESOURCE_CONN(mysql, mysql_link, MYSQLI_STATUS_VALID);
 
 
-#ifndef MYSQLI_USE_MYSQLND
-	mysql_get_character_set_info(mysql->mysql, &cs);
-	name = (char *)cs.csname;
-	collation = (char *)cs.name;
-	dir = (char *)cs.dir;
-	minlength = cs.mbminlen;
-	maxlength = cs.mbmaxlen;
-	number = cs.number;
-	state = cs.state;
-	comment = cs.comment;
-#else
 	cs = mysql->mysql->data->charset;
 	if (!cs) {
 		php_error_docref(NULL, E_WARNING, "The connection has no charset associated");
@@ -1139,7 +986,6 @@ PHP_FUNCTION(mysqli_get_charset)
 	number = cs->nr;
 	comment = cs->comment;
 	state = 1;	/* all charsets are compiled in */
-#endif
 	object_init(return_value);
 
 	add_property_string(return_value, "charset", (name) ? (char *)name : "");
@@ -1153,56 +999,6 @@ PHP_FUNCTION(mysqli_get_charset)
 }
 /* }}} */
 
-#ifndef MYSQLI_USE_MYSQLND
-extern char * mysqli_escape_string_for_tx_name_in_comment(const char * const name);
-
-/* {{{ */
-static int mysqli_begin_transaction_libmysql(MYSQL * conn, const unsigned int mode, const char * const name)
-{
-	int ret;
-	smart_str tmp_str = {0};
-	char * name_esc;
-	char * query;
-	unsigned int query_len;
-	if (mode & TRANS_START_WITH_CONSISTENT_SNAPSHOT) {
-		if (tmp_str.s) {
-			smart_str_appendl(&tmp_str, ", ", sizeof(", ") - 1);
-		}
-		smart_str_appendl(&tmp_str, "WITH CONSISTENT SNAPSHOT", sizeof("WITH CONSISTENT SNAPSHOT") - 1);
-	}
-	if (mode & TRANS_START_READ_WRITE) {
-		if (tmp_str.s) {
-			smart_str_appendl(&tmp_str, ", ", sizeof(", ") - 1);
-		}
-		smart_str_appendl(&tmp_str, "READ WRITE", sizeof("READ WRITE") - 1);
-	} else if (mode & TRANS_START_READ_ONLY) {
-		if (tmp_str.s) {
-			smart_str_appendl(&tmp_str, ", ", sizeof(", ") - 1);
-		}
-		smart_str_appendl(&tmp_str, "READ ONLY", sizeof("READ ONLY") - 1);
-	}
-	smart_str_0(&tmp_str);
-
-	name_esc = mysqli_escape_string_for_tx_name_in_comment(name);
-	query_len = spprintf(&query, 0, "START TRANSACTION%s %s",
-						 name_esc? name_esc:"", tmp_str.s? ZSTR_VAL(tmp_str.s):"");
-
-	smart_str_free(&tmp_str);
-	if (name_esc) {
-		efree(name_esc);
-	}
-
-	ret = mysql_real_query(conn, query, query_len);
-	efree(query);
-
-	if (ret && mode & (TRANS_START_READ_WRITE | TRANS_START_READ_ONLY) && mysql_errno(conn) == 1064) {
-		php_error_docref(NULL, E_WARNING, "This server version doesn't support 'READ WRITE' and 'READ ONLY'. Minimum 5.6.5 is required");
-	}
-	return ret;
-}
-/* }}} */
-#endif
-
 /* {{{ Starts a transaction */
 PHP_FUNCTION(mysqli_begin_transaction)
 {
@@ -1225,33 +1021,13 @@ PHP_FUNCTION(mysqli_begin_transaction)
 		RETURN_THROWS();
 	}
 
-#ifndef MYSQLI_USE_MYSQLND
-	if (mysqli_begin_transaction_libmysql(mysql->mysql, flags, name)) {
-		RETURN_FALSE;
-	}
-#else
 	if (FAIL == mysqlnd_begin_transaction(mysql->mysql, flags, name)) {
 		RETURN_FALSE;
 	}
-#endif
 	RETURN_TRUE;
 }
 /* }}} */
 
-#ifndef MYSQLI_USE_MYSQLND
-/* {{{ */
-static int mysqli_savepoint_libmysql(MYSQL * conn, const char * const name, bool release)
-{
-	int ret;
-	char * query;
-	unsigned int query_len = spprintf(&query, 0, release? "RELEASE SAVEPOINT `%s`":"SAVEPOINT `%s`", name);
-	ret = mysql_real_query(conn, query, query_len);
-	efree(query);
-	return ret;
-}
-/* }}} */
-#endif
-
 /* {{{ Starts a transaction */
 PHP_FUNCTION(mysqli_savepoint)
 {
@@ -1269,11 +1045,7 @@ PHP_FUNCTION(mysqli_savepoint)
 		RETURN_THROWS();
 	}
 
-#ifndef MYSQLI_USE_MYSQLND
-	if (mysqli_savepoint_libmysql(mysql->mysql, name, false)) {
-#else
 	if (FAIL == mysqlnd_savepoint(mysql->mysql, name)) {
-#endif
 		RETURN_FALSE;
 	}
 	RETURN_TRUE;
@@ -1296,11 +1068,7 @@ PHP_FUNCTION(mysqli_release_savepoint)
 		zend_argument_value_error(ERROR_ARG_POS(2), "cannot be empty");
 		RETURN_THROWS();
 	}
-#ifndef MYSQLI_USE_MYSQLND
-	if (mysqli_savepoint_libmysql(mysql->mysql, name, true)) {
-#else
 	if (FAIL == mysqlnd_release_savepoint(mysql->mysql, name)) {
-#endif
 		RETURN_FALSE;
 	}
 	RETURN_TRUE;
