@@ -87,7 +87,7 @@ static void php_pgsql_do_connect(INTERNAL_FUNCTION_PARAMETERS,int persistent);
 static void php_pgsql_get_link_info(INTERNAL_FUNCTION_PARAMETERS, int entry_type);
 static void php_pgsql_get_result_info(INTERNAL_FUNCTION_PARAMETERS, int entry_type);
 static void php_pgsql_get_field_info(INTERNAL_FUNCTION_PARAMETERS, int entry_type);
-static void php_pgsql_data_info(INTERNAL_FUNCTION_PARAMETERS, int entry_type);
+static void php_pgsql_data_info(INTERNAL_FUNCTION_PARAMETERS, int entry_type, bool nullable_row);
 static void php_pgsql_do_async(INTERNAL_FUNCTION_PARAMETERS,int entry_type);
 
 static ssize_t php_pgsql_fd_write(php_stream *stream, const char *buf, size_t count);
@@ -184,10 +184,9 @@ static const php_stream_ops php_stream_pgsql_fd_ops = {
 ZEND_BEGIN_MODULE_GLOBALS(pgsql)
 	zend_long num_links,num_persistent;
 	zend_long max_links,max_persistent;
-	bool allow_persistent;
-    int ignore_notices;
+	zend_long allow_persistent;
 	zend_long auto_reset_persistent;
-	int log_notices;
+	int ignore_notices,log_notices;
 	zend_object *default_link; /* default link when connection is omitted */
 	HashTable field_oids;
 	HashTable table_oids;
